precision mediump float;

// input from glUniform's
uniform      vec3            u_color;
uniform      float           u_alpha;
uniform      bool            u_diffuse_enabled;
uniform      vec3            u_light_dir;
uniform      float           u_light_ambiance;
uniform      bool            u_texture_enabled;
uniform      sampler2D       u_texture;
uniform      bool            u_shadow_enabled;
uniform lowp sampler2DShadow u_shadow_map;
uniform      vec2            u_uv_offset;
uniform      bool            u_uv_scale_enabled;
uniform      vec2            u_uv_scale;
uniform      bool            u_fog_enabled;
uniform      float           u_fog_distance;
uniform      vec4            u_fog_color;
uniform      float           u_fog_pow;
uniform      vec3            u_camera_pos;

// input from vertex shader
in vec3 position;       // world coordinates
in vec4 position_light; // light-space coordinates
in vec3 normal;         // world coordinates (normalized)
in vec2 uv;

// output to GPU
layout(location=0) out vec4 fragment_color;

const float pi = 3.14159265;

void main() {

#if 0 // DEBUG: visualize depth buffer
    fragment_color=vec4(vec3(gl_FragCoord.z),1.0);
    return;
#endif

#if 0 // DEBUG: visualize normals with colors
    fragment_color=vec4(normal,1.0);
    return;
#endif

    // diffuse light
    float diffuse_factor;
    if (u_diffuse_enabled) {
        diffuse_factor = max(0.0,dot(normal,-u_light_dir));
    } else {
        diffuse_factor = 1.0-u_light_ambiance;
    }

    if (u_shadow_enabled) {
        // https://learnopengl.com/Advanced-Lighting/Shadows/Shadow-Mapping
        vec3 proj_coords=position_light.xyz*0.5+0.5;
#if 0 // add bias
        proj_coords.z-=0.002;
#endif
        float shadow;
        if (proj_coords.z>0.99) {
            // no shadow outside shadow_map
            shadow=0.0;
        } else {
#if 0 // PCF for smooth edges
            // very expensive on android
            vec2 texel_size = 1.0/vec2(textureSize(u_shadow_map,0));
            // +---+---+---+---+---+
            // |   |   |   |   |   |
            // +---+---+---+---+---+
            // |   |   |   |   |   |
            // +---+---+---+---+---+
            // |-2 |-1 | X | 1 | 2 |
            // +---+---+---+---+---+
            // |   |   |   |   |   |
            // +---+---+---+---+---+
            // |   |   |   |   |   |
            // +---+---+---+---+---+
            int g=1;
            int steps=0;
            float sum=0.0;
            vec3 p=proj_coords;
            for(int x=-g;x<=g;++x) {
                for(int y=-g;y<=g;++y) {
                    p.xy=proj_coords.xy+vec2(x,y)*texel_size;
                    sum+=texture(u_shadow_map,p);
                    steps++;
                }
            }
            shadow=1.0-sum/float(steps);
#else
            shadow=1.0-texture(u_shadow_map,proj_coords);
#endif
        }
        diffuse_factor*=1.0-shadow;
    }

    float light_factor = min(1.0,u_light_ambiance+diffuse_factor);
    vec4 c = vec4(u_color*light_factor,u_alpha);

    // texture
    if (u_texture_enabled) {
        vec2 uv2 = uv+u_uv_offset;
        if (u_uv_scale_enabled) {
            uv2 *= u_uv_scale;
        }
        c *= texture(u_texture,uv2);
    }

    // fog
    if (u_fog_enabled) {
        float camera_distance = distance(u_camera_pos,position);
        float fog_factor = pow(clamp(camera_distance/u_fog_distance,0.0,1.0),u_fog_pow);
        float dy = position.y-u_camera_pos.y;
        float dx = distance(position.xz,u_camera_pos.xz);
        float fog_height = (atan(dy,dx)+pi/2.0)/pi;
        vec3 c1 = mix(c.xyz,u_fog_color.xyz,u_fog_color.w);
        c = mix(c,vec4(c1,1.0),fog_factor);
    }

    fragment_color = c;
}

